/*
 *  opn_driver.h
 *  opn_driver
 *
 *
 *  Copyright 2015 Opticon Inc. All rights reserved.
 *  Created by Nima Nourozi on 1/11/16.
 *
 */

#ifndef opn_driver_
#define opn_driver_

/***********************************************************************************************
 *	INCLUDES
 ***********************************************************************************************/
#include <CoreFoundation/CoreFoundation.h>
#include <CoreServices/CoreServices.h>
#include <map>
#include <string>
#include <vector>
#include <mutex>          
#include <iostream>
#include <chrono>
#include <thread>
#include <ctime>
#include <functional>
#include <sstream>
#include <chrono>

#include "OptiException.hpp"
#include "OpticonDataEvent.hpp"
#include "OpticonDeviceEvent.hpp"
#include "EventHandler.hpp"
#include "Event.hpp"
#include "EventBus.hpp"
#include "Barcode.hpp"

/***********************************************************************************************
 *	DEFINES
 ***********************************************************************************************/

// ERROR CODES
#define STATUS_OK						0
#define COMMUNICATIONS_ERROR			-1
#define BAD_PARAM						-2
#define SETUP_ERROR						-3
#define INVALID_COMMAND_NUMBER			-4
#define COMMAND_LRC_ERROR				-7
#define RECEIVED_CHARACTER_ERROR		-8
#define GENERAL_ERROR					-9
#define POLLING_ALREADY_ENABLED			-10
#define POLLING_ALREADY_DISABLED		-11
#define DEVICE_NOT_OPEN                 -12
#define FILE_NOT_FOUND					2
#define ACCESS_DENIED					5
#define NO_ERROR_ENCOUNTERED			6

#define DATA_AVAILABLE                  (1L)
#define NO_DATA_AVAILABLE               (0L)
#define DETERMINE_SIZE                  (0L)

/***********************************************************************************************
************************************* Parameter Options ****************************************
************************************************************************************************/
#define PARAM_GS1_DATABAR				20				// See options for PARAM_GS1_DATABAR below
#define PARAM_CODE39					31				// 0: Disable / 1: Enable
#define PARAM_UPC						9				// 0: Disable / 1: Enable
#define PARAM_CODE128					8				// 0: Disable / 1: Enable
#define PARAM_CODE39_FULL_ASCII			54				// 0: Disable / 1: Enable
#define PARAM_UPC_SUPPS					53				// See options for PARAM_UPC_SUPPS below
#define PARAM_UPCE_TO_UPCA				41				// 0: Disable / 1: Enable
#define PARAM_EAN8_TO_EAN13				42				// 0: Disable / 1: Enable
#define PARAM_EAN8_TO_EAN13_TYPE		55				// 0: Disable / 1: Enable
#define PARAM_TX_UPCA_CHECK_DIGIT		43				// 0: Disable / 1: Enable
#define PARAM_TX_UPCE_CHECK_DIGIT		44				// 0: Disable / 1: Enable
#define PARAM_CODE39_CHECK_DIGIT		46				// 0: Disable / 1: Enable
#define PARAM_TX_CODE39_CHECK_DIGIT		45				// 0: Disable / 1: Enable
#define PARAM_UPCE_PREAMBLE				37				// See options for PARAM_UPCE_PREAMBLE below
#define PARAM_UPCA_PREAMBLE				36				// See options for PARAM_UPCA_PREAMBLE below
#define PARAM_EAN128					52				// 0: Disable / 1: Enable
#define PARAM_COUPON_CODE				56				// 0: Disable / 1: Enable
#define PARAM_I2OF5						58				// 0: Disable / 1: Enable
#define PARAM_I2OF5_CHECK_DIGIT			65				// See options for PARAM_I2OF5_CHECK_DIGIT below
#define PARAM_TX_I2OF5_CHECK_DIGIT		64				// 0: Disable / 1: Enable
#define PARAM_ITF14_TO_EAN13			63				// 0: Disable / 1: Enable
#define PARAM_I2OF5_LENGTH_1			59				// Min-max: 0-30
#define PARAM_I2OF5_LENGTH_2			60				// Min-max: 0-30
#define PARAM_D2OF5						57				// 0: Disable / 1: Enable
#define PARAM_D2OF5_LENGTH_1			61				// Min-max: 0-30
#define PARAM_D2OF5_LENGTH_2			62				// Min-max: 0-30
#define PARAM_UPC_EAN_SECURITY			47				// See options for PARAM_UPC_EAN_SECURITY below
#define PARAM_UPC_EAN_REDUNDANCY		48				// See options for PARAM_UPC_EAN_REDUNDANCY below
#define PARAM_SCANNER_ON_TIME			17				// Min-max: 10-100 (1 to 10 seconds)
#define PARAM_BUZZER_VOLUME				2				// See options for PARAM_BUZZER_VOLUME below
#define PARAM_COMM_AWAKE_TIME			32				// Min-max: 1-6
#define PARAM_BAUDRATE					13				// See options for PARAM_BAUDRATE below
#define PARAM_BAUDRATE_SWITCH_DELAY		29				// Min-max: 0-100
#define PARAM_RESET_BAUDRATES			28				// 0: Disable / 1: Enable
#define PARAM_REJECT_REDUNDANT_CODES	4				// See options for PARAM_REJECT_REDUNDANT_CODES below
#define PARAM_HOST_CONNECT_BEEP			10				// 0: Disable / 1: Enable
#define PARAM_HOST_COMPLETE_BEEP		11				// 0: Disable / 1: Enable
#define PARAM_BATTERY_LOW_INDICATION	7				// See options for PARAM_BATTERY_LOW_INDICATION below
#define PARAM_AUTO_CLEAR				15				// 0: Disable / 1: Enable
#define PARAM_DELETE_ENABLE				33				// See options for PARAM_DELETE_ENABLE below
#define PARAM_DATA_PROTECTION			49				// 0: Disable / 1: Enable
#define PARAM_MEMORY_FULL_INDICATION	50				// 0: Disable / 1: Enable
#define PARAM_MEMORY_INDICATION			51				// See options for PARAM_MEMORY_INDICATION below
#define PARAM_MAX_BARCODE_LENGTH		34				// Max barcode length changed from 30 to 254 (255=unlimited) for PX20 (2D symbologies)
#define PARAM_GOODREAD_LED_ONTIME		30				// See options for PARAM_GOODREAD_LED_ONTIME below
#define PARAM_STORE_RTC					35				// 0: Disable / 1: Enable
#define PARAM_ASCII_MODE				79				// 0: Disable / 1: Enable
#define PARAM_BEEPER_TOGGLE				85				// 0: Disable / 1: Enable
#define PARAM_BEEPER_AUTO_ON			86				// 0: Disable / 1: Enable
#define PARAM_LASER_MODE				5				// See options for PARAM_LASER_MODE below
#define PARAM_AIMING_ON_TIME			19				// See options for PARAM_AIMING_ON_TIME below
#define PARAM_SCRATCH_PAD				38				// 32 byte scratch pad

// General Enable / disable parameter options
#define PARAM_DISABLE					0
#define PARAM_ENABLE					1

// General Off / On paramter options
#define PARAM_OFF						0
#define PARAM_ON						1

// Options for PARAM_UPC_SUPPS
#define UPC_NO_SUPPLEMENTS				0
#define UPC_SUPPLEMENTS_ONLY			1
#define UPC_SUPPLEMENTS_AUTO			2

// Options for PARAM_UPCE_PREAMBLE / PARAM_UPCA_PREAMBLE
#define UPC_PREAMBLE_NONE				0
#define UPC_PREAMBLE_SYSTEM				1
#define UPC_PREAMBLE_SYSTEM_N_COUNTRY	2

// Options for PARAM_I2OF5_CHECK_DIGIT
#define I2OF5_CHECK_DIGIT_DISABLED		0
#define I2OF5_CHECK_DIGIT_USS			1
#define I2OF5_CHECK_DIGIT_OPCC			2

// Options for PARAM_BATTERY_LOW_INDICATION (bit flags)
#define BATTLOW_NO_OPERATION_NO_INDICATION		0
#define BATTLOW_INDICATION_ENABLED				1
#define BATTLOW_ALLOW_OPERATION					2

// Options for PARAM_DELETE_ENABLE (bit flags)
#define CLEAR_ALL_ENABLED						1			// Allow clearing of all data by pressing delete key for 6 seconds
#define DELETE_ENABLED							2			// Allow deleting of a barcode from memory using the delete key
#define SET_DEFAULT_ENABLED						4			// Allow reset to factory default by pressing boths keys for 10 seconds

// Options for PARAM_GOODREAD_LED_ONTIME
#define GOODREAD_ONTIME_250MS					1
#define GOODREAD_ONTIME_500MS					2
#define GOODREAD_ONTIME_750MS					3
#define GOODREAD_ONTIME_1000MS					4

// Options for PARAM_GS1_DATABAR (bit flags)
#define GS1_DATABAR_DISABLED					0
#define GS1_DATABAR_LIMITED						1
#define GS1_DATABAR_14							2
#define GS1_DATABAR_EXPANDED					4

// Options for PARAM_LASER_MODE
#define LASER_MODE_NORMAL						0
#define LASER_MODE_AIMING_HELP					1
#define LASER_MODE_SMALL						2

// Options for PARAM_REJECT_REDUNDANT_CODES
#define REJECT_REDUNDANT_CODES_OFF				0
#define REJECT_CONSECUTIVE_REDUNDANT_CODES		1
#define REJECT_ANY_REDUNDANT_CODES				2

// Options for PARAM_BUZZER_VOLUME
#define BUZZER_DISABLED							0
#define BUZZER_MAXIMUM							1
#define BUZZER_LOUD								2
#define BUZZER_NORMAL							3
#define BUZZER_SOFT								4
#define BUZZER_MINIMUM							5

// Options for PARAM_MEMORY_INDICATION
#define MEMORY_INDICATION_DISABLED				0
#define MEMORY_INDICATION_90PCT					1
#define MEMORY_INDICATION_NOT_EMPTY				2


// Global event type variables
static const UInt32				kEventClassOpticon = 'OPTI';
static const UInt32				kEventKindInsert = 1;
static const UInt32				kEventKindRemove = 2;
static const UInt32				kEventKindData = 3;

#pragma GCC visibility push(default)
/***********************************************************************************************
 *	CLASS DECLARATION
 ***********************************************************************************************/
class OPN_Interface
{

    //
    ///***********************************************************************************************
    // *	API
    // ***********************************************************************************************/
    //
public:
    
    /*!
     @function OPN_Interface
     @abstract Constructor: creates a new OPN_Interface object
     */
    OPN_Interface() throw();
    
    
    /*!
     @function ~OPN_Interface
     @abstract Destructor: called when an OPN_Interface object is destroyed
     */
    ~OPN_Interface();
    
    
    /*!
     @function Initialize
     @abstract Initializes the interface internal variables and objects.
     @discussion ***IMPORTANT*** This function has to be called before any other functions to
     ensure the interface's internals have been instantiated.
     @return 1 to indicate success -1 if the initialization failed.
     */
    int Initialize();
    
    
    /*!
     @function addDeviceEventListener
     @abstract Registers an Opticon device listener class.
     @param an event handler object to receive usb events
     @return A pointer to a HandlerRegistration class that can be used to deregister the registered class
     @discussion In order to receive notificaitions when an Opticon device is connected/disconnected
     the user has to register an event handler class.
     */
    HandlerRegistration* addDeviceEventListener(EventHandler<OpticonDeviceEvent> & handler);
    

    /*!
     @function addDataEventListener
     @abstract Registers a data listener class.
     @param an event handler object to receive data events
     @return A pointer to a HandlerRegistration class that can be used to deregister the registered class
     @discussion In order to receive notificaitions of new data the user has to register an event handler class.
     A notification is sent when there is data and the device is connected or polling is enabled.
     */
    HandlerRegistration* addDataEventListener(EventHandler<OpticonDataEvent> & handler);
    
    
    /*!
     @function removeDeviceEventListener
     @abstract Deregisters a device listener class.
     @param an event handler object to receive data events
     @discussion All registered event listener classes should be removed when not needed any longer to prevent memory leaks.
     */
    void removeDeviceEventListener(HandlerRegistration* handler);
    
    
    /*!
     @function removeDataEventListener
     @abstract Deregisters a data listener class.
     @param an event handler object to receive data events
     @discussion All registered event listener classes should be removed when not needed any longer to prevent memory leaks.
     */
    void removeDataEventListener(HandlerRegistration* handler);

    
    /*!
     @function getTime
     @abstract Returns the device's current timestamp
     @param deviceHandler   USB device to send the command to
     @param timestamp       timestamp from the device
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown.
     */
    SInt32 getTime(UInt32 deviceHandler, struct tm& timestamp);
    
    
    /*!
     @function setTime
     @abstract Set the device's timestamp to the timestamp passed in
     @param deviceHandler   USB device to send the command to
     @param timestamp       new timestamp
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown.
     */
    SInt32 setTime(UInt32 deviceHandler, struct tm& timestamp);
    
    
    /*!
     @function interrogate
     @abstract Check the status of the device. Also retrieves the device ID and software version.
     @param deviceHandler   USB device to send the command to
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The interrogate command
     requests a response from the CS-1504 unit indicating that it is operational. The CS-1504 may not correctly
     receive the interrogate command sent from the Host. Erroneous data received or a response time-out prior to
     a confirmed interrogate shall be ignored by both the CS-1504 unit and Host.
     After a successful call to interrogate user can retrieve the device ID and the software version.
     */
    SInt32 interrogate(UInt32 deviceHandler);
    
    
    /*!
     @function powerDown
     @abstract Powers down the device
     @param deviceHandler   USB device to send the command to
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The power down transaction shall 
     direct the CS-1504 unit to end the communication session and enter power down mode until the next scan, function
     or HOST_IN activation. The CS-1504 unit shall send the response message, initiate an external signal, and enter 
     power down mode.
     */
    SInt32 powerDown(UInt32 deviceHandler);
    
    
    /*!
     @function clearData
     @abstract Clears all data stored on the device
     @param deviceHandler   USB device to send the command to
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The clear bar codes 
     command shall request the CS-1504 unit to clear all stored bar codes from its memory.
     */
    SInt32 clearData(UInt32 deviceHandler);
    
    
    /*!
     @function setDefaults
     @abstract Restore default settings
     @param deviceHandler   USB device to send the command to
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The set defaults transaction 
     shall cause all the user parameter values to be set to their default values. In addition, all stored bar codes 
     will be cleared from the CS-1504 memory.
     */
    SInt32 setDefaults(UInt32 deviceHandler);
    
    /*!
     @function getParam
     @abstract Return the requested parameter setting value from the device
     @param deviceHandler   USB device to send the command to
     @param paramValue      parameter value returned from the device
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The getParam command
     shall request download of parameter data to the CS-1504 unit.     
     */
    SInt32 getParam(UInt32 deviceHandler, UInt8 Setting, int& ParamValue);
    
    
    /*!
     @function setParam
     @abstract Set the new parameter setting on the device
     @param deviceHandler   USB device to send the command to
     @param paramValue      new parameter value to sent to the device
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The setParam command
     shall download the new parameter data to the CS-1504 unit.
     */
    SInt32 setParam(UInt32 deviceHandler, UInt8 Setting, int ParamValue);
    
    
    /*!
     @function isDataAvailable
     @abstract Checks if there is any (stale) data since last call to getAllBarcodes.
     @param deviceHandler   USB device to send the command to
     @param retVal          boolean value returned from the device
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The isDataAvailable command
     returns true if there is any data stored on the CS-1504 unit. Note this returns any data that has been read
     since last call to getAllBarcodes. In other words it returns stale data.
     */
    SInt32 isDataAvailable(UInt32 deviceHandler, bool& retVal);
    
    
    /*!
     @function getAllBarcodes
     @abstract Returns all the barcode data from the device
     @param deviceHandler   USB device to send the command to
     @param barcodeList     list to copy the barcodes returned from the device to
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The getAllBarcodes command
     returns a list of barcodes currently on the CS-1504 unit. If there are no data an empty list is returned.
     */
    SInt32 getAllBarcodes(UInt32 deviceHandler, std::vector<Barcode> & barcodeList);
    
    
    /*!
     @function getBarcodeCount
     @abstract Returns number of barcodes stored on the device
     @param deviceHandler   USB device to send the command to
     @param count           nubmer of barcodes available
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The getBarcodeCount command
     returns number of barcodes currently on the CS-1504 unit.
     */
    UInt32 getBarcodeCount(UInt32 deviceHandler, int & count);
    
    
    /*!
     @function getProtocolVersion
     @abstract Returns the protocol version from the device
     @param deviceHandler   USB device to send the command to
     @param PV              protocol version of the device
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The getProtocolVersion command
     returns the protocol version from the CS-1504 unit.
     */
    SInt32 getProtocolVersion(UInt32 deviceHandler, UInt32& PV);
    
    
    /*!
     @function getDeviceID
     @abstract Returns the device ID of the device
     @param deviceHandler   USB device to send the command to
     @param PV              Device ID of the device
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The getDeviceID command
     returns the device ID from the CS-1504 unit.
     */
    SInt32 getDeviceID(UInt32 deviceHandler, std::string& deviceId);
    
    
    /*!
     @function getSystemStatus
     @abstract Returns the devices status
     @param deviceHandler   USB device to send the command to
     @param SysStat         Status of the device
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The getSystemStatus command
     returns the status from the CS-1504 unit.
     */
    SInt32 getSystemStatus(UInt32 deviceHandler, UInt32& SysStat);
    
    
    /*!
     @function getSoftwareVersion
     @abstract Returns the software version of the device
     @param deviceHandler   USB device to send the command to
     @param SysStat         Current software version of the device
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The getSoftwareVersion command
     returns the software version from the CS-1504 unit.
     */
    SInt32 getSoftwareVersion(UInt32 deviceHandler, std::string& version);
    
    
    /*!
     @function getRetryCount
     @abstract Returns current retries
     @param Retries         Current retries
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion Returns the current retries used in case the device did not respond to a command.
     */
    SInt32 getRetryCount(UInt32& Retries);
    
    
    /*!
     @function setRetryCount
     @abstract Set the current retries
     @param Retries         New retries value
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion Set the retries used in case the device did not respond to a command.
     */
    SInt32 setRetryCount(UInt32 Retries);
    
    /*!
     @function getASCIIMode
     @abstract Returns the packet mode of the device. Packets are in ASCII mode or Binary mode.
     @param deviceHandler   USB device to send the command to
     @param AM              ASCII mode setting
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The getASCIIMode command
     returns the ASCII mode from the CS-1504 unit. Packets are in ASCII mode or Binary mode.
     */
    SInt32 getASCIIMode(UInt32 deviceHandler, bool& AM);
    
    
    /*!
     @function getASCIIMode
     @abstract Returns the RTC setting of the device.
     @param deviceHandler   USB device to send the command to
     @param AM              RTC setting
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion A valid deviceHandler has to passed in otherwise an error is thrown. The getRTCMode command
     returns the RTC setting from the CS-1504 unit. If the RTC mode is set barcode data sent from the device
     will include timestamp data.
     */
    SInt32 getRTCMode(UInt32 deviceHandler, bool& RTC);
    
    
    /*!
     @function getLibraryVersion
     @abstract Returns the software version for the library
     @param VersionString         Software version
     @return STATUS_OK if successful. Otherwise an error code.
     @discussion NONE
     */
    SInt32 getLibraryVersion(std::string& VersionString);


    
/***********************************************************************************************
 *	PRIVATE VARIABLES
 ***********************************************************************************************/
private:
    class               impl;
    std::auto_ptr<impl> pimpl;
    SInt32 setTime(UInt32 deviceHandler);

};

/***********************************************************************************************
 *	CONSTRUCTOR AND DESTRUCTOR TO LOAD THE DYLIB IN USER APPLICATION
 ***********************************************************************************************/
extern "C" OPN_Interface* NewOPN_Interface(void);
typedef OPN_Interface * OPN_Interface_creator(void);
extern "C" void DeleteOPN_Interface(OPN_Interface* driver);
typedef void OPN_Interface_disposer(OPN_Interface* driver);

#pragma GCC visibility pop
#endif
